#include "class.h"

#include "../tools/misc.h"
#include "../tools/class_is_derived_from.h"
#include "../ast_traverser/definition_finder.h"
#include "design.h"

risc::Class::Class(SgClassDefinition *ast_node, Type type):
  Definition(ast_node, type)
{
  initialize_variables();
}

risc::Class::~Class()
{
  for(VariableVectorIter
      iter  = variables_.begin();
      iter != variables_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(PrimitiveChannelInstanceVectorIter
      iter  = sub_prim_channel_.begin();
      iter != sub_prim_channel_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(HierarchicalChannelInstanceVectorIter
      iter  = sub_hier_channel_.begin();
      iter != sub_hier_channel_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(ModuleInstanceVectorIter
      iter  = sub_modules_.begin();
      iter != sub_modules_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(FunctionVectorIter
      iter  = functions_.begin();
      iter != functions_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(EventVectorIter
      iter  = events_.begin();
      iter != events_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(InPortVectorIter
      iter  = inports_.begin();
      iter != inports_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(InOutPortVectorIter
      iter  = inoutports_.begin();
      iter != inoutports_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(OutPortVectorIter
      iter  = outports_.begin();
      iter != outports_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }
}

void
risc::Class::initialize_functions()
{
  // get ctor of the class (module)
  SgMemberFunctionDeclaration *ctor =
    SageInterface::getDefaultConstructor(get_ast_node()->get_declaration());

  // get dtor of the class (module)
  SgMemberFunctionDeclaration *dtor =
    SageInterface::getDefaultDestructor(get_ast_node()->get_declaration());

  // Find all functions in the class defintion
  Rose_STL_Container<SgNode*> function_definitions =
    NodeQuery::querySubTree(get_ast_node(), V_SgFunctionDefinition);

  // The constructor is in the function_definitions container because rose
  // treats a constructor as a special kind of a function. So, we have to
  // search for the ctor and delete it.
  //TODO
  SgFunctionDefinition *ctor_definition =
    risc::traverser::get_function_definition(ctor);

  function_definitions.erase(
    std::remove(function_definitions.begin(), function_definitions.end(),
             ctor->get_definition()),
    function_definitions.end());

  // same for the destructor
  if(dtor) {
    function_definitions.erase(
      std::remove(function_definitions.begin(), function_definitions.end(),
               dtor->get_definition()),
      function_definitions.end());
  }

}

SgClassDefinition*
risc::Class::get_ast_node()
{
  return isSgClassDefinition(ast_node_);
}

std::string
risc::Class::get_ast_type_name()
{
  return get_ast_node()->get_qualified_name().getString();
}

std::string risc::Class::get_name()
{
  return SageInterface::get_name(get_ast_node());
}

void
risc::Class::initialize_variables()
{
  // This loop iterates all members of the class. A member can be a
  // constructor, typedef, function, and so on. For this case we have to
  // distinguish the different types
  for(SgDeclarationStatementPtrList::iterator
      iter  = get_ast_node()->get_members().begin();
      iter != get_ast_node()->get_members().end();
      iter++) {

    SgVariableDeclaration *vd = isSgVariableDeclaration(*iter);

    // We are only interested in variable declarations
    if(!vd)
      continue;

    SgVariableDefinition *var_def = vd->get_definition();

    SgType* type = var_def->get_type();

    if(isSgReferenceType(type)) {

      type = isSgReferenceType(type)->get_base_type();
    }

    // perhaps we have an array of int's, signals, ports, events, or
    // prim. channels
    if(isSgArrayType(type)) {
      type = SageInterface::getArrayElementType(type);

      SgClassType* class_type =
        isSgClassType(SageInterface::getArrayElementType(type));
      if(class_type) {

        SgClassDeclaration* class_decl =
          isSgClassDeclaration(class_type->get_declaration());
      }
    }

    // We check if it is a SystemC element
    //
    // A sc_module instance has two significant attributes:
    //  a) the instance is a data structure (class)
    //  b) the class is derived from sc_module
    //
    // A port
    //  a) derived from sc_port
    if(risc::tools::is_data_structure(type)) {

      SgClassType *ct = isSgClassType(type);

      SgClassDeclaration *declaration =
        isSgClassDeclaration(ct->get_declaration());

      if(declaration->isForward()) {
        declaration =
          isSgClassDeclaration(declaration->get_definingDeclaration());
      }

      SgClassDefinition *definition = declaration->get_definition();

      //Okay, at this point we know it is a C++ data structure.

      //case of a module
      if(risc::tools::is_derived_from_sc_module(definition) ) {

        // At this point, we have the SgClassDefinition of the corresponding
        // class member variable in the module. In the next step, we have to
        // create a Module instance and insert into the design

        Module *md =
          GlobalDesign::design_static->find_module_definition(definition);

        sub_modules_.push_back(new ModuleInstance(var_def, md));
        continue;
      }

      // case of a port
      if(risc::tools::is_derived_from_sc_port_base(definition) ) {

        SgTemplateInstantiationDefn* templ_def =
          isSgTemplateInstantiationDefn(definition);
        if(templ_def) {

          // case of an in_port
          if(risc::tools::is_sc_in_port(templ_def)){

            inports_.push_back(new InPort(var_def));
            continue;
          }

          // case of an out_port
          if(risc::tools::is_sc_out_port(templ_def)){

            outports_.push_back(new OutPort(var_def));
            continue;
          }

          // case of an inout_port
          if(risc::tools::is_sc_inout_port(templ_def)){

            inoutports_.push_back(new InOutPort(var_def));
            continue;
          }
        }

        SgClassDefinition *class_def = isSgClassDefinition(definition);
        if(class_def) {

          // case of an sc_in_clk
          if(risc::tools::is_sc_in_clk(class_def)){

            inports_.push_back(new InPort(var_def));
            continue;
          }

          // case of an sc_out_clk
          if(risc::tools::is_sc_out_clk(class_def)){

            outports_.push_back(new OutPort(var_def));
            continue;
          }

          // case of an sc_inout_clk
          if(risc::tools::is_sc_inout_clk(class_def)){

            inoutports_.push_back(new InOutPort(var_def));
            continue;
          }
        }

        // in any other case we can not define it. So it must be an inout port
        inoutports_.push_back(new InOutPort(var_def));
        continue;
      }

      // case of a event
      if(risc::tools::is_sc_event(definition)) {

        events_.push_back(new Event(var_def));
        continue;
      }

      // case of a primitive channel
      if(risc::tools::is_derived_from_sc_prim_channel(definition)) {

        sub_prim_channel_.push_back(new PrimitiveChannelInstance(var_def,NULL));
        continue;
      }

      // case of a channel (this case represents a hierarchical channel)
      if(risc::tools::is_derived_from_sc_channel(definition)) {

        // FIXME
        // A hierarchical has to have a user definition!
        // This is only for test cases
        sub_hier_channel_.push_back(
                              new HierarchicalChannelInstance(var_def,NULL));
        continue;
      }
    }

    // We check if it is an int, double, float, unsigned, ...
    if(risc::tools::is_builtin_type(type) ||
       risc::tools::is_data_structure(type)) {

      variables_.push_back(new Variable(var_def));
    }

    if(SageInterface::isReferenceType(var_def->get_type())) {

      variables_.push_back(new Variable(var_def));
    }
  }
}

risc::Port*
risc::Class::find_port(SgVariableDefinition* var_def)
{
  for(InPortVectorIter
      iter  = inports_.begin();
      iter != inports_.end();
      iter++) {
    if(var_def == (*iter)->ast_node_)
      return *iter;
  }

  for(InOutPortVectorIter
      iter  = inoutports_.begin();
      iter != inoutports_.end();
      iter++) {
    if(var_def == (*iter)->ast_node_)
      return *iter;
  }

  for(OutPortVectorIter
      iter  = outports_.begin();
      iter != outports_.end();
      iter++) {
    if(var_def == (*iter)->ast_node_)
      return *iter;
  }
  return NULL;
}

risc::Event*
risc::Class::find_event(SgVariableDefinition* var_def)
{
  for(EventVectorIter iter = events_.begin(); iter != events_.end(); iter++) {
    if(var_def == (*iter)->ast_node_)
      return *iter;
  }
  return NULL;
}

risc::Variable*
risc::Class::find_variable(SgVariableDefinition* var_def)
{
  for(VariableVectorIter
      iter  = variables_.begin();
      iter != variables_.end();
      iter++) {

    if(var_def == (*iter)->ast_node_)
      return *iter;
  }
  return NULL;
}

SgFunctionDefinition*
risc::Class::get_before_end_of_elaboration_definition()
{

  for(SgDeclarationStatementPtrList::iterator
      iter  = get_ast_node()->get_members().begin();
      iter != get_ast_node()->get_members().end();
      iter++) {

    if(SageInterface::get_name(*iter) == "before_end_of_elaboration") {
      SgFunctionDeclaration *func_decl = isSgFunctionDeclaration(*iter);
      func_decl = isSgFunctionDeclaration(func_decl->get_definingDeclaration());

      return func_decl->get_definition();
    }
  }

  return NULL;
}

risc::PrimitiveChannelInstance*
risc::Class::find_prim_channel(SgVariableDefinition* var_def)
{
  for(PrimitiveChannelInstanceVectorIter
      iter  = sub_prim_channel_.begin();
      iter != sub_prim_channel_.end();
      iter++) {

    if(var_def == (*iter)->ast_node_)
      return *iter;
  }

  return NULL;
}

risc::HierarchicalChannelInstance*
risc::Class::find_hier_channel(SgVariableDefinition* var_def)
{
  for(HierarchicalChannelInstanceVectorIter
      iter  = sub_hier_channel_.begin();
      iter != sub_hier_channel_.end();
      iter++) {

    if(var_def == (*iter)->ast_node_)
      return *iter;
  }

  return NULL;
}

risc::ModuleInstance*
risc::Class::find_module(SgVariableDefinition* var_def)
{
  for(ModuleInstanceVectorIter
      iter  = sub_modules_.begin();
      iter != sub_modules_.end();
      iter++) {

    if(var_def == (*iter)->ast_node_)
      return *iter;
  }

  return NULL;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
