#include "design.h"

#include <vector>

#include "../ast_traverser/module_traverser.h"
#include "../ast_traverser/find_sc_main.h"
#include "../ast_traverser/path_checker.h"
#include "../tools/misc.h"
#include "../tools/command_line_parsing.h"
#include "../tools/search_for_typedef_declaration.cpp"
#include "../tools/file_classifier.h"

#include "module_instance.h"

risc::Design::Design(int argc, char* argv[]):
  path_instance_mapper_()
{
  old_argc_ = argc;
  old_argv_ = argv;

  std::cout << "Command line old: "
            << CommandlineProcessing::generateStringFromArgList(
                 CommandlineProcessing::generateArgListFromArgcArgv(argc, argv))
            << std::endl;

  std::vector<std::string> new_arguments
    = risc::tools::preparsing_commandline_arguments(argc, argv, input_files_);

  std::cout << "Command line new: "
            << CommandlineProcessing::generateStringFromArgList(
                 CommandlineProcessing::generateArgListFromArgcArgv(argc, argv))
            << std::endl;

  //project_ = frontend(argc, argv);
  project_ = frontend(old_argc_, old_argv_);

  if(project_ == NULL) {
    exit(1);
  }

  systemc_definitions_ =
    risc::tools::SystemCDefinitionFinder::get_definitions(project_);

  risc::GlobalDesign::design_static = this;

  main_ = findMain(project_);
  sc_main_ = traverser::find_sc_main(project_);

  initialize_defintions();
  initialize_top_modules();
  initialize_global_variables();
  //initialize_global_functions();

  // This must be called last because the we need some members of the design
  path_instance_mapper_.determine_lookup_table(get_root());
}

risc::Design::~Design()
{
  // clear all vectors

  for(ModuleVectorIter iter = module_definitions_.begin();
      iter != module_definitions_.end(); iter++) {
    delete *iter;
  }

  for(PrimitiveChannelVectorIter
      iter = primitive_channel_.begin();
      iter != primitive_channel_.end(); iter++) {
    delete *iter;
  }

  for(HierarchicalChannelVectorIter
      iter  = hierarchical_channel_.begin();
      iter != hierarchical_channel_.end(); iter++) {
    delete *iter;
  }

  for(EventVectorIter iter = events_.begin(); iter != events_.end(); iter++ ) {
    delete *iter;
  }

  for(FunctionVectorIter iter = functions_.begin(); iter != functions_.end();
      iter++) {
    delete *iter;
  }
}

void risc::Design::initialize_defintions()
{
  // Search for all class definitions that are derived from sc_module and
  // are not defined in a file in the include path
  std::vector<SgClassDefinition*> modules = risc::traverser::find_all_modules();

  for(std::vector<SgClassDefinition*>::iterator
      iter  = modules.begin();
      iter != modules.end();
      iter++) {

    SgClassDefinition* temp_cd = isSgClassDefinition(*iter);

    if(temp_cd) {

      module_definitions_.push_back(new Module(temp_cd));
    }
  }

  //FIXME Has to continue with channels
}

risc::Module* risc::Design::find_module_definition(SgClassDefinition* cd)
{
  Module* temp = NULL;

  for(ModuleVectorIter iter = module_definitions_.begin() ;
      iter != module_definitions_.end(); iter++) {

    if(risc::tools::class_types_are_identical((*iter)->get_ast_node(),cd)) {
      temp = *iter;
    }
  }

  return temp;
}

void risc::Design::initialize_top_modules()
{
  //FIXME get sc_main function and search for members that are variable
  //      instances and derived from sc_module

  SgFunctionDefinition* func_def = sc_main_->get_definition();
  SgBasicBlock* bb = func_def->get_body();

  for(Rose_STL_Container<SgStatement*>::iterator
      iter  = bb->get_statements().begin();
      iter != bb->get_statements().end();
      iter++) {

    SgVariableDeclaration* var_decl =  isSgVariableDeclaration(*iter);

    if(var_decl) {

      SgType* type = var_decl->get_decl_item(0)->get_type();

      // in this case we are addressing an int, float, ...
      if(tools::is_builtin_type(type))
        continue;

      SgArrayType* array_type = isSgArrayType(type);

      // in this case we are addressing a array of int, float, ...
      if(array_type)
        if(tools::is_builtin_type(SageInterface::getArrayElementType(type)))
          continue;

      // At this point we can know that it is a class or a corresponding array
      SgClassDefinition* class_def = NULL;

      // If it is an array we have to get the definition.
      // So, we are asking for the definiton behind the array
      if(array_type) {

        SgClassType* class_type =
          isSgClassType(SageInterface::getArrayElementType(type));


        SgClassDeclaration* class_decl =
          isSgClassDeclaration(class_type->get_declaration());

        class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

        class_def = class_decl->get_definition();

      } else {
        class_def = tools::class_definition_of_variable(var_decl);
      }

      if(tools::is_derived_from_sc_module(class_def)) {

        Module* md = find_module_definition(class_def);

        top_modules_.push_back(
                         new ModuleInstance(var_decl->get_definition(), md));
      }
    }
  }
}

void risc::Design::initialize_global_variables()
{
  SgDeclarationStatementPtrList &global_decls
    = SageInterface::getGlobalScope(sc_main_)->get_declarations();

  int counter = 0;

  for(SgDeclarationStatementPtrList::iterator
      iter  = global_decls.begin();
      iter != global_decls.end();
      iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter);

    if(var_decl) {

      // check if we added the variable already
      bool exists_already = false;
      for(VariableVectorIter
          iter  = global_variables_.begin();
          iter != global_variables_.end();
          iter++) {

        if(var_decl->get_decl_item(0)->search_for_symbol_from_symbol_table()
            == (*iter)->get_ast_node()->get_declaration()->
                   get_decl_item(0)->search_for_symbol_from_symbol_table()) {
          exists_already = true;
        }
      }

      if(exists_already) {
        continue;
      }

      global_variables_.push_back(new Variable(var_decl->get_definition()));
    }
  }
}

void risc::Design::initialize_global_functions()
{
  Rose_STL_Container<SgNode*> function_definition
    = NodeQuery::querySubTree(project_, V_SgFunctionDefinition);

  int filter = 0;
  for(Rose_STL_Container<SgNode*>::const_iterator
      i  = function_definition.begin();
      i != function_definition.end(); ++i) {

    SgFunctionDefinition *func = isSgFunctionDefinition(*i);

    // Theses functions are not in the global scope
    if(SageInterface::getScope(func->get_declaration()) !=
        SageInterface::getGlobalScope(func->get_declaration())) {
      continue;
    }

    // Skip all template functions
    if(isSgTemplateFunctionDefinition(func) ||
        isSgTemplateInstantiationFunctionDecl(func->get_declaration()) ||
        isSgTemplateInstantiationMemberFunctionDecl(func->get_declaration())) {
      continue;
    }

    if(func->get_file_info()->isCompilerGenerated()) {
      continue;
    }

    // Skip all functions which are part of a class
    if(func->get_declaration()->get_specialFunctionModifier().isConstructor() ||
        func->get_declaration()->get_specialFunctionModifier().isDestructor() ||
        isSgMemberFunctionDeclaration(func->get_declaration())) {
      continue;
    }

    // These functions are built in functions. So, we can skipt them
    if(risc::tools::is_builtin_function(func->get_declaration())) {
      continue;
    }

    // Global functions are not allowed to start with a '_'
    // https://gcc.gnu.org/onlinedocs/libstdc++/manual/source_code_style.html
    if(SageInterface::get_name(func).at(0) == '_' ) {
        //&& SageInterface::get_name(func).at(1) == '_'){
      continue;
    }

    if(SageInterface::get_name(func) == "sc_main")
      continue;

    global_functions_.push_back(new Function(func));
  }
}

void risc::Design::print_design()
{
  std::cout << "Modules" << std::endl;
  for(risc::ModuleInstanceVectorIter
      iter  = top_modules_.begin();
      iter != top_modules_.end(); iter++) {

    std::cout << " " << (*iter)->get_ast_type_name()
              << " "
              << (*iter)->get_name()
              << std::endl;

    (*iter)->definition_->print_hierarchy(2);
  }
}

void risc::Design::print_all_class_definitions()
{
  for(ModuleVectorConstIter iter = module_definitions_.begin();
      iter != module_definitions_.end(); iter++) {

    std::cout << (*iter)->get_ast_type_name() << std::endl;
  }
}

void risc::Design::print_all_global_functions()
{
  std::cout << std::endl << "Global functions" << std::endl;
  for(risc::FunctionVectorConstIter
      iter  = global_functions_.begin();
      iter != global_functions_.end(); iter++) {

    std::cout << " " << (*iter)->get_name() << std::endl;
  }
}

risc::Design* risc::GlobalDesign::design_static = NULL;

risc::Design* risc::get_design()
{
  return risc::GlobalDesign::design_static;
}

SgProject* risc::get_project()
{
  return risc::GlobalDesign::design_static->project_;
}

risc::Variable*
risc::Design::find_global_variable(SgVariableDefinition* var_def)
{
  for(VariableVectorIter
      iter  = global_variables_.begin();
      iter != global_variables_.end();
      iter++) {

    if(var_def == (*iter)->ast_node_)
      return *iter;
  }
  return NULL;
}

risc::InstanceTree
risc::Design::get_root()
{
  InstanceTree result;

  if(top_modules_.size() != 1) {
    std::cout << "Number of modules: " << top_modules_.size() << std::endl;
    assert(false && "Only exactly one top module is allowed");
  }

  InstanceTree::Element top(NULL, top_modules_.at(0));
  result.path_.push_back(top);

  return result;
}

void
risc::Design::reset_instance_counter()
{
  for(ModuleVectorIter
      iter  = module_definitions_.begin();
      iter != module_definitions_.end();
      iter++) {
    (*iter)->instance_counter_ = 0;
  }

  for(HierarchicalChannelVectorIter
      iter  = hierarchical_channel_.begin();
      iter != hierarchical_channel_.end();
      iter++) {
    (*iter)->instance_counter_ = 0;
  }
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
