#ifndef SEGMENT_GRAPH_H_INCLUDED_
#define SEGMENT_GRAPH_H_INCLUDED_

#include<iostream>
#include<set>
#include<utility>

#include<boost/graph/adjacency_list.hpp>
#include<boost/graph/graphviz.hpp>
#include<boost/algorithm/string.hpp>

#include "../internal_representation/function.h"
#include "segment.h"
#include "edge.h"

namespace risc {

namespace sg {

class SegmentGraph;

typedef boost::adjacency_list<boost::listS,
        boost::listS, boost::bidirectionalS, Segment, Edge,
        boost::property<boost::vertex_index_t, int> > Graph;
typedef boost::graph_traits<Graph>::vertex_descriptor VertexDescriptor;
typedef boost::graph_traits<Graph>::edge_descriptor EdgeDescriptor;

typedef boost::graph_traits<risc::sg::Graph>::vertex_iterator VertexIterator;

class CachedFunctionAstAttributes;

/**
 *  \class SegmentGraph
 *  \brief This class represents a segment graph for a process
 *
 */
class SegmentGraph {

public:
  typedef std::set<SgFunctionDefinition*> FunctionCallBoundarySet;
  typedef std::set<VariantT> CCxxKeywordsBoundarySet;

  SegmentGraph(FunctionCallBoundarySet function_boundaries);
  SegmentGraph(CCxxKeywordsBoundarySet keyword_boundaries);
  SegmentGraph(FunctionCallBoundarySet function_boundaries,
      CCxxKeywordsBoundarySet keyword_boundaries);

  /**
   * \function clean_graph
   * \brief    This function deletes all duplicated nodes in each vertex
   *           of the graph
   */
  void clean_graph();


  /**
   * \function print_graph
   * \brief    This function prints the graph in a dot file.
   *           The name of the file is user defined by argument
   * \param    filename   The name of the file
   */
  void print_graph(std::string filename);


  /**
   * \function print_graph_read_write_access
   * \brief    This function prints all variables which have read or written
   *           in this segment.
   * \param    filename   The name of the file
   */
  void print_graph_read_write_access(std::string filename);


  /**
   * \function add_function
   * \brief    This function analyzes the given function and creates a
   *           segment graph for it. The graph will be automatically inserted
   *           in the existing graph. The flag duplicate_segments will be
   *           used to generate an instance segement graph.
   */
  void add_function(Function *function, bool duplicate_segments = false);

  typedef std::set<VertexDescriptor> SegmentSet;

  /**
   * \function read_write_analysis_of_segments
   * \brief    The function determines which variables are read and written in
   *           the individual segments
   */
  void read_write_analysis_of_segments();

  /**
   * \function set_all_segments_to_untouched
   * \brief    This function sets the status of the flag
   *           helper_for_graph_algorithms for all nodes to 'Untouched'
   */
  void set_all_segments_to_untouched();

private:

  /**
   * \function  handle_recursive_calls
   * \brief     This function integrates the recursive function calls
   */
  void handle_recursive_calls();


  /**
   * \function  separate_variable_declaration_and_initializer_on_demand
   * \brief     This function checks if a direct or indirect boundary call is
   *            in the initializer expression of the variable declaration.
   *            If yes, the declaration and initialization will be separated.
   *            Example:
   *            Input:
   *                int x = y + foo(); // foo() has a indirect boundary
   *            Output:
   *                int x;
   *                x = y + foo();
   */
  void separate_variable_declaration_and_initializer_on_demand(
           SgBasicBlock *&bb, bool duplicate_segments);


  /**
   * \function  decompose_expression_with_boundary_calls
   * \brief     This function decomposes an expression into multiple
   *            statements. Functions which call directly or indirectly a
   *            boundary will be separated and have temporally variables.
   *            Example:
   *            Input:
   *                x = y + foo(); // foo() has a indirect boundary
   *            Output:
   *                int temp_foo = foo();
   *                int x = y + temp_foo();
   */
  void decompose_expression_with_boundary_calls(SgExpression *expr,
           SgBasicBlock *&bb, bool duplicate_segments);


  /**
   * \function create_temp_variable_for_expression
   * \brief    This function creates a variable declaration of the type of the
   *           given expression. The variable will be inserted before the
   *           associated expression statement of the given expression.
   */
  void create_temp_variable_for_expression(SgExpression *expr,
           SgBasicBlock *&bb);


  /**
   * \function decompose_member_function_calls
   * \brief    This function decomposes member function calls like in the
   *           following given example:
   *           m + x.y->foo() + n; // where foo() has a boundary statement
   *             into
   *           Y *y = x.y;
   *           int temp_foo;
   *           temp_foo = y->foo()
   *           m + temp_foo + n;
   *
   *           It works for the dot operator with references as well.
   */
  SgExpression* decompose_member_function_calls(SgFunctionCallExp *func_call,
                    SgBasicBlock *&bb);


  /**
   * \function generate_unique_name
   * \brief    Generates an unique name for a varible based on the given node.
   *
   */
  std::string generate_unique_name(SgNode *node);


  /**
   * \function   is_boundary_stmt
   * \brief      This function checks if a given statement is a potential
   *             boundary statement
   * \param      current_stmt The statement which will be checked
   */
  bool is_boundary_stmt(SgStatement const * const current_stmt);


  /**
   * \function has_function_call_with_boundary
   * \brief    This function checks if in the given expression a function
   *           call occurs which calls a functon with a boundary statements.
   */
  SgFunctionCallExp* has_function_call_with_boundary(SgExpression *expr,
                         bool duplicate_segments);


  /**
   * \function should_decompose_function
   * \brief    This function determines if a expression should be decomposed.
   *           This is the case if expression has the following structure:
   *           var = somepointer->func();
   *           var = somevariable.func();
   *           var = func();
   *           and var is a global, member, or static variable.
   */
  bool should_decompose_function(SgExpression *expr);


  /**
   * \function contains_call_expression
   * \brief    In this function we look for expressions which contain a function
   *           call in the following style:
   *           foo();
   *           this->foo();
   *           var = foo();
   *           var = this->foo();
   *           The return value will be the function declaration of the called
   *           function
   */
  SgFunctionDeclaration* contains_function_call_expression(SgExpression *expr);


  /**
   * \function build_graph
   * \brief    This function creates the segment graph
   *
   * \param    current_stmt        The first statement of the sequence which
   *                               will be analyzed
   * \param    current_segments    The beginning segments
   */
  SegmentSet build_graph(
      SgStatement *current_stmt,
      SegmentSet current_segments,
      SegmentSet &break_segments,
      SegmentSet &continue_segments,
      bool duplicate_segments);


  /**
   * \function add_expression_to_segment
   * \brief    This function adds to each segment the given expression
   *
   * \param    segments   These segments which will contain the given
   *                      expression
   * \param    expr       The expressions which will added to all segments
   */
  void add_expression_to_segment(SegmentSet segments, SgNode *expr);


  /**
   * \function build_segment_graph_for_function
   * \brief    This function generates a segment graph for a function
   */
  CachedFunctionAstAttributes* build_segment_graph_for_function(
      SgFunctionDefinition *func_def, bool duplicate_segments);


  /**
   * \function insert_loop_edges
   * \brief    This is a helper function for loop handling
   *
   */
  void insert_loop_edges(VertexDescriptor &loop_vertex,
           SegmentSet &current_segments,
           SegmentSet &continue_segments,
           SegmentSet &leaf_segments_of_loop);

  SegmentSet duplicate_segments(SegmentSet segments);
  SegmentSet duplicate_empty_segments(SegmentSet segments);

  FunctionCallBoundarySet function_boundaries_;
  CCxxKeywordsBoundarySet keyword_boundaries_;

  std::set<VertexDescriptor> recursive_function_calls_;

  static int counter;
public:
  Graph graph_;

  std::set<VertexDescriptor> precached_function_segments_;
};

} // end of namespace sg

} // end of namespace risc

#endif /* SEGMENT_GRAPH_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
